import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from basicsr.models.archs.arch_util import LayerNorm2d



class SimpleGate(nn.Module):
    def forward(self, x, y):
        return x * y

class ProjUnit(nn.Module):
    def __init__(self, c):
        super().__init__()
        self.conv = nn.Sequential(nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0, groups=1, bias=True),
                                  nn.Conv2d(c, c, kernel_size=3, stride=1, padding=1, groups=c, bias=True)
                                  )
    def forward(self, x):
        x = self.conv(x)
        return  x

class ProjUnitC(nn.Module):
    def __init__(self, c):
        super().__init__()
        self.conv = nn.Conv2d(c, c, kernel_size=1, stride=1, padding=0, groups=1, bias=True)

    def forward(self, x):
        x = self.conv(x)
        return  x

class ProjUnitT(nn.Module):
    def __init__(self, c):
        super().__init__()
        self.conv = nn.Conv2d(c, c, kernel_size=3, stride=1, padding=1, groups=c, bias=False)

    def forward(self, x):
        x = self.conv(x)
        return  x

class MultiGatedBlock(nn.Module):
    def __init__(self, c, drop_out_rate=0.):
        super().__init__()
        self.sg = SimpleGate()

        self.norm = LayerNorm2d(c)
        self.proj_h = ProjUnit(c)
        self.proj_r = ProjUnitC(c)
        self.proj_z = ProjUnitT(c)
        self.proj_t = ProjUnitC(c)

        self.dropout1 = nn.Dropout(drop_out_rate) if drop_out_rate > 0. else nn.Identity()
        self.beta = nn.Parameter(torch.zeros((1, c, 1, 1)), requires_grad=True)

    def forward(self, inp):
        x = inp
        x = self.norm(x)

        h = self.proj_h(x)
        r = self.proj_r(x)
        z = self.proj_z(x)

        ht = self.sg(h, r)
        ht = self.proj_t(ht)
        ht = self.sg(ht, 1-z)
        h = self.sg(h, z)

        out = inp + (ht + h) * self.beta

        return out

class MG_MLP(nn.Module):
    def __init__(self, img_channel=3, width=16, middle_blk_num=1, enc_blk_nums=[], dec_blk_nums=[]):
        super().__init__()

        self.intro = nn.Conv2d(in_channels=img_channel, out_channels=width, kernel_size=3, padding=1, stride=1, groups=1, bias=True)
        self.ending = nn.Conv2d(in_channels=width, out_channels=img_channel, kernel_size=3, padding=1, stride=1, groups=1, bias=True)
        self.encoders = nn.ModuleList()
        self.decoders = nn.ModuleList()
        self.middle_blks = nn.ModuleList()
        self.ups = nn.ModuleList()
        self.downs = nn.ModuleList()

        chan = width
        for i, num in enumerate(enc_blk_nums):
            self.encoders.append(
                nn.Sequential(
                    *[MultiGatedBlock(chan) for _ in range(num)]
                )
            )
            self.downs.append(
                nn.Conv2d(chan, 2*chan, 2, 2)
            )
            chan = chan * 2

        self.middle_blks = \
            nn.Sequential(
                *[MultiGatedBlock(chan) for _ in range(middle_blk_num)]
            )

        for num in dec_blk_nums:
            self.ups.append(
                nn.Sequential(
                    nn.Conv2d(chan, chan * 2, 1, bias=False),
                    nn.PixelShuffle(2)
                )
            )
            chan = chan // 2
            self.decoders.append(
                nn.Sequential(
                    *[MultiGatedBlock(chan) for _ in range(num)]
                )
            )

        self.padder_size = 2 ** len(self.encoders)

    def forward(self, inp):
        B, C, H, W = inp.shape
        inp = self.check_image_size(inp)

        x = self.intro(inp)

        encs = []

        for encoder, down in zip(self.encoders, self.downs):
            x = encoder(x)
            encs.append(x)
            x = down(x)

        x = self.middle_blks(x)

        for decoder, up, enc_skip in zip(self.decoders, self.ups, encs[::-1]):
            x = up(x)
            x = x + enc_skip
            x = decoder(x)

        x = self.ending(x)
        x = x + inp

        return x[:, :, :H, :W]

    def check_image_size(self, x):
        _, _, h, w = x.size()
        mod_pad_h = (self.padder_size - h % self.padder_size) % self.padder_size
        mod_pad_w = (self.padder_size - w % self.padder_size) % self.padder_size
        x = F.pad(x, (0, mod_pad_w, 0, mod_pad_h))
        return x




